'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var PublicKey = require('./PublicKey.cjs');
var bytes = require('./utils/bytes.cjs');
var SdkError = require('./errors/SdkError.cjs');

class GpaBuilder {
  constructor(context, programId, fields, options = {}) {
    this.context = context;
    this.programId = programId;
    this.fields = fields;
    this.options = options;
  }
  reset() {
    return new GpaBuilder(this.context, this.programId, this.fields, {});
  }
  slice(offset, length) {
    return new GpaBuilder(this.context, this.programId, this.fields, {
      ...this.options,
      dataSlice: {
        offset,
        length
      }
    });
  }
  withoutData() {
    return this.slice(0, 0);
  }
  addFilter(...filters) {
    return new GpaBuilder(this.context, this.programId, this.fields, {
      ...this.options,
      filters: [...(this.options.filters ?? []), ...filters]
    });
  }
  where(offset, data) {
    let bytes$1;
    if (typeof data === 'string') {
      bytes$1 = bytes.base58.serialize(data);
    } else if (typeof data === 'number' || typeof data === 'bigint' || typeof data === 'boolean') {
      bytes$1 = bytes.base10.serialize(BigInt(data).toString());
    } else if (PublicKey.isPublicKey(data)) {
      bytes$1 = new Uint8Array(data.bytes);
    } else {
      bytes$1 = new Uint8Array(data);
    }
    return this.addFilter({
      memcmp: {
        offset,
        bytes: bytes$1
      }
    });
  }
  whereField(field, data) {
    if (!this.fields) {
      throw new SdkError.SdkError('Fields are not defined in this GpaBuilder.');
    }
    const fieldIndex = this.fields.findIndex(([name]) => name === field);
    if (fieldIndex < 0) {
      throw new SdkError.SdkError(`Field [${field}] is not defined in this GpaBuilder.`);
    }
    const [, serializer] = this.fields[fieldIndex];
    const offset = this.fields.slice(0, fieldIndex).reduce((acc, [, s]) => acc === null || s.fixedSize === null ? null : acc + s.fixedSize, 0);
    if (offset === null) {
      throw new SdkError.SdkError(`Field [${field}] is not in the fixed part of ` + `the account's data. In other words, it is located after ` + `a field of variable length which means we cannot find a ` + `fixed offset for the filter.`);
    }
    return this.where(offset, serializer.serialize(data));
  }
  whereSize(dataSize) {
    return this.addFilter({
      dataSize
    });
  }
  sortUsing(callback) {
    return new GpaBuilder(this.context, this.programId, this.fields, {
      ...this.options,
      sortCallback: callback
    });
  }
  async get(options = {}) {
    const accounts = await this.context.rpc.getProgramAccounts(this.programId, {
      ...options,
      dataSlice: options.dataSlice ?? this.options.dataSlice,
      filters: [...(options.filters ?? []), ...(this.options.filters ?? [])]
    });
    if (this.options.sortCallback) {
      accounts.sort(this.options.sortCallback);
    }
    return accounts;
  }
  async getAndMap(callback, options = {}) {
    return (await this.get(options)).map(callback);
  }
  async getPublicKeys(options = {}) {
    return this.getAndMap(account => account.publicKey, options);
  }
  async getDataAsPublicKeys(options = {}) {
    return this.getAndMap(account => {
      try {
        return PublicKey.publicKey(account.data);
      } catch (error) {
        const message = `Following a getProgramAccount call, you are trying to use an ` + `account's data (or a slice of it) as a public key. ` + `However, we encountered an account ` + `[${PublicKey.base58PublicKey(account.publicKey)}] whose data ` + `[base64=${bytes.base64.deserialize(account.data)}] ` + `is not a valid public key.`;
        throw new SdkError.SdkError(message);
      }
    }, options);
  }
}
const gpaBuilder = (context, programId, fields) => new GpaBuilder(context, programId, fields);

exports.GpaBuilder = GpaBuilder;
exports.gpaBuilder = gpaBuilder;
//# sourceMappingURL=GpaBuilder.cjs.map
